/* ----------------------------------------------------------------------------
 * file: W7XDbNamedObject.java
 * project: ConfigDB-BusinessSchema
 * by:  Josef Maier, Marc Lewerentz, Anett Spring
 *
 * Copyright (c) 2004 - 2014, All rights reserved.
 * Max-Planck-Institut für Plasmaphysik. W7-X CoDaC group.
 * ----------------------------------------------------------------------------
 */
package de.mpg.ipp.codac.business.schema.testcase;

import java.io.Serializable;
//import java.text.DateFormat;
//import java.text.SimpleDateFormat;
//import java.util.GregorianCalendar;
//import java.util.Objects;
import java.util.StringTokenizer;

import javax.jdo.annotations.Discriminator;
import javax.jdo.annotations.DiscriminatorStrategy;
import javax.jdo.annotations.Extension;
import javax.jdo.annotations.FetchGroup;
import javax.jdo.annotations.FetchGroups;
import javax.jdo.annotations.IdGeneratorStrategy;
import javax.jdo.annotations.IdentityType;
import javax.jdo.annotations.Inheritance;
import javax.jdo.annotations.InheritanceStrategy;
import javax.jdo.annotations.PersistenceCapable;
import javax.jdo.annotations.Persistent;
import javax.jdo.annotations.PrimaryKey;

//import de.mpg.ipp.codac.business.common.MetaInfo;
//import de.mpg.ipp.codac.business.common.ObjectID;
import de.mpg.ipp.codac.business.schema.common.W7XDbNamedObject.Id;

@PersistenceCapable(detachable = "true",
                    identityType = IdentityType.APPLICATION,
                    objectIdClass = Id.class)
@Discriminator(strategy = DiscriminatorStrategy.CLASS_NAME)
@Inheritance(strategy = InheritanceStrategy.COMPLETE_TABLE)
@FetchGroups({
    @FetchGroup(name = "PKonly", members = {
        @Persistent(name = "sysId"),
        @Persistent(name = "className") }),
    @FetchGroup(name = "MetaInfo", members = {
        @Persistent(name = "sysId"),
        @Persistent(name = "className"),
        @Persistent(name = "sysName"),
        @Persistent(name = "sysText"),
        @Persistent(name = "sysState"),
        @Persistent(name = "sysOwner"),
        @Persistent(name = "sysEditInfo"),
        @Persistent(name = "sysFlags"),
        @Persistent(name = "sysVersionInfo") }) })

public abstract class W7XDbNamedObject
    implements Serializable {

    /**
     * The Constant METAATTRIBUTES.
     * 
     * @tts Preliminary String constant to define the MetaInfo attributes for queries (because I'm
     *      not able to retrieve the members of a statically defined fetchgroup at runtime). The
     *      order is important!
     */
    public static final String METAATTRIBUTES =
        "sysId, className, sysName, sysText, sysState, sysOwner, sysEditInfo, sysFlags, sysVersionInfo ";

    /**
     * @author marcl
     * 
     */
    public static class Id
        implements Serializable {

        private static final long serialVersionUID = -3410881584863686048L;
        public Integer sysId;
        public String className;

        /**
         * Instantiates a new id.
         */
        public Id() {}

        /**
         * Instantiates a new id.
         * 
         * @param str
         *            the str
         */
        public Id( String str ) {
            StringTokenizer token = new StringTokenizer(str, ".");
            this.className = token.nextToken();
            this.sysId = Integer.valueOf(token.nextToken());
        }

        @Override
        public String toString() {
            return this.className + "." + this.sysId;
        }

        @Override
        // TODO equality by sysId and className - does this make sense?
        // Have to have in mind when adding W7xDbNamedObjects to Sets etc.!
        public boolean equals( Object o ) {
            if (this == o) {
                return true;
            }
            if (o == null) {
                return false;
            }
            if (o.getClass() != getClass()) {
                return false;
            }
            Id objToCompare = (Id) o;
            return ((this.sysId.equals(objToCompare.sysId)) && (this.className == null
                ? objToCompare.className == null
                : this.className.equals(objToCompare.className)));
        }

        @Override
        public int hashCode() {
            return this.sysId ^ this.className.hashCode();
        }

    }

    static final long serialVersionUID = 0;
    /** constants for sysFlags */
    private static final int PROTECTED = 1;
    @SuppressWarnings("unused")
    private static final int INPOOL = 2;
    @SuppressWarnings("unused")
    private static final int DELETEFROMPOOL = 4;
    private static final int NOTFORREUSE = 16;

    /**
     * Numeric part of the identifier for the W7XDbNamedObject. (Actual classname is the other
     * part.) This identifier is set (by XDV software) when a new W7XDbNamedObject is created in the
     * configuration database; it is not the objectivity ObjectId. In the (measured-data-)
     * archive(s) it only acts as reference back to the object in the configuration database. For
     * archived signals without associated configuration object (for example, data not measured by
     * CoDaC systems), sysId can be zero.
     */
    @PrimaryKey
    @Persistent(valueStrategy = IdGeneratorStrategy.INCREMENT, extensions = {
        @Extension(vendorName = "datanucleus", key = "key-cache-size", value = "1"),
        @Extension(vendorName = "datanucleus", key = "key-initial-value", value = "10000"),
        @Extension(vendorName = "datanucleus", key = "strategy-when-notnull", value = "false") })
    //    @Index(name = "sysIdIdx") // does not help in concrete classes!?
    private Integer sysId = null;

    @PrimaryKey
    private String className = this.getClass().getSimpleName();

    /** name set by XDV */
    @Persistent
    //    @Index(name = "sysNameIdx") // does not help in concrete classes!?
    protected String sysName = "";
    /** text, contains comment set by XDV */
    @Persistent
    protected String sysText = "";
    /** state, contains status information set by XDV */
    @Persistent
    protected int sysState = 0;
    /**
     * owner set by XDV; this field contains the sysId of the UserGroup and determines the access
     * rights for this object
     */
    @Persistent
    protected int sysOwner = 0;

    /**
     * The sysEditInfo field contains information set by XDV-Editor. This field is amended by
     * objyaccess write methods This field is typically used to store user-name and timestamp of the
     * last change to this object or its related configuration-database object.
     */
    @Persistent
    protected String sysEditInfo = "";

    /**
     * Zusammenfassung der Bits in sysFlags:
     * 
     * Bit 0 - PROTECT (Objektinstanz darf im Arbeitsspeicher/Datenbank nicht gelöscht werden)
     * 
     * Bit 1 - INPOOL (Eine Instanz des Objekts soll im Arbeitsspeicher angelegt werden)
     * 
     * Bit 2 - DELETEFROMPOOL (Eine Instanz des Objekts im Arbeitsspeicher kann gelöscht werden)
     * 
     * Bit 3 - Reseviert für EXTRA_FEASIBILITY_CHECK( braucht die FCS, ist noch nicht ganz
     * durchgestaltet )
     * 
     * Bit 4 - NOT_FOR_REUSE (keine Wiederverwendung erlaubt)
     * 
     * Bit 5 .. Bit 30 - können noch mit einer Bedeutung belegt werden
     * 
     * Bit 31 - IS_DUMMY (Wird von der Datenbankschnittstelle FCS <-> josDaemon verwendet)
     * 
     */
    @Persistent
    protected int sysFlags = 0;

    /**
     * The sysVersionInfo field contains information set by XDV-objyaccess. This field is set by
     * objyaccess write methods This field is compared on objyaccess modify methods to support
     * optimistic locking This field is typically used to store user-name, timestamp and version
     * number of the last change to this object or its related configuration-database object.
     */
    @Persistent
    protected String sysVersionInfo = "";

    /**
     * Field access method to set persistent field id.
     * 
     * @param id
     *            the new sys id
     * 
     *            usually sysId is set by DB - use {@link #resetSysId()} to clear
     * 
     *            for some admin processes this is still needed
     */

    public void setSysId( int id ) {
        this.sysId = id;
    }

    /**
     * Field access method to set persistent field id.
     * 
     * @param id
     *            the new sys id
     * 
     *            usually sysId is set by DB - use {@link #resetSysId()} to clear
     * 
     *            for some admin processes this is still needed
     */
    public void setSysId( Integer id ) {
        this.sysId = id;
    }

    /**
     * Field access method to get persistent field id.
     * 
     * @return the sys id
     */
    public int getSysId() {
        return sysId == null ? 0 : this.sysId;
    }

    /**
     * Reset sysId - to force generating a new ID when storing to DB. Used to mark an object as new.
     * 
     */
    public void resetSysId() {
        sysId = null;
    }

  
    /**
     * Field access method to set persistent field name.
     * 
     * @param name
     *            the new sys name
     */
    public void setSysName( String name ) {
        this.sysName = name;
    }

    /**
     * Field access method to get persistent field name.
     * 
     * @return the sys name
     */
    public String getSysName() {
        return this.sysName;
    }

    /**
     * Field access method to set persistent field text.
     * 
     * @param text
     *            the new sys text
     */
    public void setSysText( String text ) {
        this.sysText = text;
    }

    /**
     * Field access method to get persistent field text.
     * 
     * @return the sys text
     */
    public String getSysText() {
        return this.sysText;
    }

    /**
     * Field access method to set persistent field state.
     * 
     * @param state
     *            the new sys state
     */
    public void setSysState( int state ) {
        this.sysState = state;
    }

    /**
     * Reset sysState to 0.
     * 
     */
    public void resetSysState() {
        this.sysState = 0;
    }

    /**
     * <br>
     * Field access method to get persistent field state.
     * 
     * @return the sys state
     */
    public int getSysState() {
        return this.sysState;
    }

    /**
     * <br>
     * Field access method to set persistent field owner.
     * 
     * @param owner
     *            the new sys owner
     */
    public void setSysOwner( int owner ) {
        this.sysOwner = owner;
    }

    /**
     * <br>
     * Field access method to get persistent field owner.
     * 
     * @return the sys owner
     */
    public int getSysOwner() {
        return this.sysOwner;
    }


    /**
     * Reset sysEditInfo to emtpy string.
     */
    public void resetSysEditInfo() {
        sysEditInfo = "";
    }


    /**
     * Gets the sysEditInfo (unformatted).
     * 
     * @return the sysEditInfo
     */
    public String getSysEditInfo() {
        return sysEditInfo;
    }

    /**
     * Sets the sysEditInfo.
     * 
     * @param sysEditInfo
     *            the new sysEditInfo
     */
    public void setSysEditInfo( String sysEditInfo ) {
        this.sysEditInfo = sysEditInfo;
    }

    /**
     * <br>
     * Field access method to set persistent field sysFlags.
     * 
     * @param flags
     *            the new sys flags
     */
    public void setSysFlags( int flags ) {
        this.sysFlags = flags;
    }

    /**
     * <br>
     * Field access method to get persistent field sysFlags.
     * 
     * @return the sys flags
     */
    public int getSysFlags() {
        return this.sysFlags;
    }

    /**
     * <br>
     * Field access method to set/reset persistent field PROTECTED in sysFlags.
     * 
     * @param b
     *            the new protected
     */
    public void setProtected( boolean b ) {
        if (b)
            this.sysFlags |= PROTECTED;
        else
            this.sysFlags &= ~PROTECTED;
    }

    /**
     * <br>
     * Field access method to test persistent field PROTECTED in sysFlags.
     * 
     * @return true, if is protected
     */
    public boolean isProtected() {
        if ((this.sysFlags & PROTECTED) == 0)
            return false;
        else
            return true;
    }

    /**
     * <br>
     * Field access method to set/reset persistent field NOTFORREUSE in sysFlags.
     * 
     * @param b
     *            the new notforreuse
     */
    public void setNotforreuse( boolean b ) {
        if (b)
            this.sysFlags |= NOTFORREUSE;
        else
            this.sysFlags &= ~NOTFORREUSE;
    }

    /**
     * <br>
     * Field access method to test persistent field PROTECTED in sysFlags.
     * 
     * @return true, if is notforreuse
     */
    public boolean isNotforreuse() {
        if ((this.sysFlags & NOTFORREUSE) == 0)
            return false;
        else
            return true;
    }


    /**
     * Sets the default.
     */
    public void setDefault() {
        /*
         * !Do not reset any sysXXX parameters because they contain meta information
         * about the object not to be deleted or reset!
         */
    }

    /**
     * Gets the object info in the form "className.idNum - objectName".
     * 
     * @return the object info as "className.idNum - objectName"
     */
    public String getObjectInfo() {
        String classnameLong, classname;
        classnameLong = this.getClass().getName();
        int pos1 = classnameLong.lastIndexOf('.');
        int pos2 = classnameLong.lastIndexOf(";");
        if (pos2 > pos1)
            classname = classnameLong.substring(pos1 + 1, pos2);
        else
            classname = classnameLong.substring(pos1 + 1);
        return classname + "." + this.getSysId() + " - " + this.getSysName();
    }

    /**
     * Gets the objects version info for optimistic locking.
     * 
     * @return the object version info as String
     */
    public String getSysVersionInfo() {
        if (this.sysVersionInfo == null)
            return "no sysVersionInfo available";
        return this.sysVersionInfo;
    }


    @Override
    public String toString() {
        return this.getClass().getSimpleName() + "." + getSysId();
    }

    public String toNameString() {
        return this.getSysName() + " [" + toString() + "]";
    }


    /**
     * Sets the sys version info.
     * 
     * @param sysVersionInfo
     *            the new sys version info
     */
    public final void setSysVersionInfo( String sysVersionInfo ) {
        this.sysVersionInfo = sysVersionInfo;
    }

    /**
     * Gets the class name.
     * 
     * @return the class name
     */
    public String getClassName() {
        return className;
    }

}
